import torch
import torch.nn as nn
import dgl.function as fn
# from dgl.base import DGLError
from dgl.utils import expand_as_pair


def has_self_loop(graph):
    return any(graph.edges()[0] - graph.edges()[1] == 0)


class NGCFConv(nn.Module):
    def __init__(self, in_features, out_features, activation=None):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.activation = activation
        self.weight_self = nn.Parameter(torch.empty(in_features, out_features))
        self.weight_interaction = nn.Parameter(
            torch.empty(in_features, out_features))
        nn.init.xavier_uniform_(self.weight_self)
        nn.init.xavier_uniform_(self.weight_interaction)

    def forward(self, graph, feature):
        with graph.local_scope():
            # TODO very slowly
            # if has_self_loop(graph):
            #     raise DGLError(
            #         'You should first remove self loop with `dgl.remove_self_loop`.'
            #     )

            feat_src, feat_dst = expand_as_pair(feature, graph)

            degs = graph.out_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1, ) * (feat_src.dim() - 1)
            norm = torch.reshape(norm, shp)
            feat_src = feat_src * norm

            graph.srcdata['h'] = feat_src
            graph.dstdata['h'] = feat_dst

            graph.update_all(fn.copy_u('h', 'm'), fn.sum('m', 'h_self'))

            def message_func(edges):
                return {'m': torch.mul(edges.src['h'], edges.dst['h'])}

            graph.update_all(message_func, fn.sum('m', 'h_interaction'))

            rst = torch.matmul(
                graph.dstdata['h_self'], self.weight_self) + torch.matmul(
                    graph.dstdata['h_interaction'], self.weight_interaction)
            # TODO adding the interaction part will reduce the performance

            degs = graph.in_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1, ) * (feat_dst.dim() - 1)
            norm = torch.reshape(norm, shp)
            rst = rst * norm

            if self.activation is not None:
                rst = self.activation(rst)

            return rst

    def extra_repr(self):
        """Set the extra representation of the module,
        which will come into effect when printing the model.
        """
        return f'in={self.in_features}, out={self.out_features}, activation={self.activation}'


if __name__ == '__main__':
    import dgl
    graph = dgl.graph(([0, 0, 1, 2, 2], [1, 2, 2, 1, 0]))
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    graph = graph.to(device)
    model = NGCFConv(2, 2).to(device)
    inputs = torch.tensor([
        [0.2, 0.5],
        [0.6, 0.4],
        [0.3, 0.3],
    ]).to(device)
    print(model(graph, inputs))



import numpy as np
from sklearn.metrics import roc_auc_score, ndcg_score
from multiprocessing import Pool


def recall(y_trues, y_scores, k):
    assert y_trues.shape == y_scores.shape
    assert len(y_trues.shape) == 2
    orders = np.argsort(y_scores, axis=-1)[:, ::-1][:, :k]
    return np.mean(
        np.sum(np.take_along_axis(y_trues, orders, axis=-1), axis=-1) /
        np.sum(y_trues, axis=-1))


def mrr(y_trues, y_scores):
    assert y_trues.shape == y_scores.shape
    assert len(y_trues.shape) == 2
    orders = np.argsort(y_scores, axis=-1)[:, ::-1]
    y_trues = np.take_along_axis(y_trues, orders, axis=-1)
    rr_scores = y_trues / (np.arange(y_trues.shape[1]) + 1)
    return np.mean(np.sum(rr_scores, axis=-1) / np.sum(y_trues, axis=-1))


def fast_roc_auc_score(y_trues, y_scores, num_processes=None):
    # TODO can it be faster?
    with Pool(processes=num_processes) as pool:
        return np.mean(pool.starmap(roc_auc_score, zip(y_trues, y_scores)))



import torch
import torch.nn as nn

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


class NCF(nn.Module):
    def __init__(self, args, graph, first_num, second_num):
        super().__init__()
        self.args = args
        # Load graph for negative sampling
        self.graph = graph
        self.primary_etypes = [
            x for x in graph.canonical_etypes if not x[1].endswith('-by')
        ]
        assert args.non_graph_embedding_dim % 4 == 0
        # TODO put the following dims into parameters
        self.first_embedding = nn.ModuleDict({
            'GMF':
            nn.Embedding(first_num, args.non_graph_embedding_dim // 4),
            'MLP':
            nn.Embedding(first_num, args.non_graph_embedding_dim)
        })
        self.second_embedding = nn.ModuleDict({
            'GMF':
            nn.Embedding(second_num, args.non_graph_embedding_dim // 4),
            'MLP':
            nn.Embedding(second_num, args.non_graph_embedding_dim),
        })
        self.MLP = nn.Sequential(
            nn.Linear(2 * args.non_graph_embedding_dim,
                      args.non_graph_embedding_dim),
            nn.ReLU(),
            nn.Linear(args.non_graph_embedding_dim,
                      args.non_graph_embedding_dim // 2),
            nn.ReLU(),
            nn.Linear(args.non_graph_embedding_dim // 2,
                      args.non_graph_embedding_dim // 4),
        )
        self.final_linear = nn.Linear(args.non_graph_embedding_dim // 2,
                                      1,
                                      bias=False)

    def forward(self, first, second, task_name=None, provided_embeddings=None):
        '''
        Args:
            first: {
                'name': str,
                'index': (shape) batch_size
            },
            second: {
                'name': str,
                'index': (shape) batch_size
            }
        '''
        first_index = first['index']
        second_index = second['index']

        # batch_size, non_graph_embedding_dim // 4
        GMF_vector = torch.mul(
            self.first_embedding['GMF'](first_index),
            self.second_embedding['GMF'](second_index),
        )
        # batch_size, non_graph_embedding_dim // 4
        MLP_vector = self.MLP(
            torch.cat((
                self.first_embedding['MLP'](first_index),
                self.second_embedding['MLP'](second_index),
            ),
                      dim=-1))
        # batch_size
        score = self.final_linear(torch.cat((GMF_vector, MLP_vector),
                                            dim=-1)).squeeze(dim=1)
        return score



import torch.nn as nn
import dgl
from dgl.nn.pytorch import GraphConv

from .base import HeterogeneousAggregator


class GCN(HeterogeneousAggregator):
    def __init__(self, graph_embedding_dims, etypes):
        super().__init__(graph_embedding_dims, etypes)

    def get_layer(self, input_dim, output_dim, current_layer, total_layer):
        assert 0 <= current_layer <= total_layer - 1
        if current_layer < total_layer - 1:
            return GraphConv(input_dim, output_dim, activation=nn.ReLU())

        return GraphConv(input_dim, output_dim)


if __name__ == '__main__':
    import torch
    from ....utils import add_reverse
    graph = dgl.heterograph(
        add_reverse({
            ('user', 'follow', 'category'):
            (torch.randint(10, (100, )), torch.randint(10, (100, ))),
            ('user', 'buy', 'game'):
            (torch.randint(10, (100, )), torch.randint(10, (100, ))),
            ('user', 'play', 'game'): (torch.randint(10, (100, )),
                                       torch.randint(10, (100, ))),
            ('game', 'belong', 'category'): (torch.randint(10, (100, )),
                                             torch.randint(10, (100, ))),
        }), {
            'user': 10,
            'category': 10,
            'game': 10
        })
    model = GCN([16, 12, 8, 6], graph.canonical_etypes)
    print(model)

    # Test full graph
    input_embeddings1 = {
        node_name: torch.rand(10, 16)
        for node_name in ['user', 'category', 'game']
    }
    input_embeddings2 = {
        etype:
        {node_name: torch.rand(10, 16)
         for node_name in [etype[0], etype[2]]}
        for etype in graph.canonical_etypes if not etype[1].endswith('-by')
    }
    for x in [input_embeddings1, input_embeddings2]:
        outputs = model([graph] * 3, x)
        for k, v in outputs.items():
            for k2, v2 in v.items():
                print(k, k2, v2.shape)

    # Test mini-batch
    eid_dict = {
        etype: graph.edges('eid', etype=etype)
        for etype in graph.canonical_etypes
    }
    dataloader = dgl.dataloading.EdgeDataLoader(
        graph,
        eid_dict,
        dgl.dataloading.MultiLayerFullNeighborSampler(3),
        batch_size=4,
        shuffle=True,
    )
    for input_nodes, _, blocks in dataloader:
        input_embeddings1 = {
            k: torch.rand(v.size(0), 16)
            for k, v in input_nodes.items()
        }
        input_embeddings2 = {
            etype: {
                k: torch.rand(v.size(0), 16)
                for k, v in input_nodes.items() if k in [etype[0], etype[2]]
            }
            for etype in graph.canonical_etypes if not etype[1].endswith('-by')
        }

        for x in [input_embeddings1, input_embeddings2]:
            outputs = model(blocks, x)
            for k, v in outputs.items():
                for k2, v2 in v.items():
                    print(k, k2, v2.shape)

        break



import torch.nn as nn


class BPRLoss(nn.Module):
    def __init__(self):
        super().__init__()
        raise NotImplementedError

    def forward(self, input, target):
        pass


class MarginLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, positive, negative):
        return (1 - negative.view(positive.shape[0], -1) +
                positive.unsqueeze(1)).clamp(min=0).mean()



import torch.nn as nn
import dgl
from dgl.nn.pytorch import HeteroGraphConv
# TODO add dropout?


class HeterogeneousAggregator(nn.Module):
    def __init__(self, graph_embedding_dims, etypes):
        super().__init__()
        self.primary_etypes = [x for x in etypes if not x[1].endswith('-by')]
        assert len(graph_embedding_dims) >= 2
        self.layer_dict = nn.ModuleDict()
        for etype in self.primary_etypes:
            layers = nn.ModuleList()
            for i in range(len(graph_embedding_dims) - 2):
                layers.append(
                    HeteroGraphConv({
                        e: self.get_layer(graph_embedding_dims[i],
                                          graph_embedding_dims[i + 1], i,
                                          len(graph_embedding_dims) - 1)
                        for e in [etype[1], f'{etype[1]}-by']
                    }))
            layers.append(
                HeteroGraphConv({
                    e: self.get_layer(graph_embedding_dims[-2],
                                      graph_embedding_dims[-1],
                                      len(graph_embedding_dims) - 2,
                                      len(graph_embedding_dims) - 1)
                    for e in [etype[1], f'{etype[1]}-by']
                }))
            self.layer_dict[str(etype)] = layers

    def get_layer(self, input_dim, output_dim, current_layer, total_layer):
        raise NotImplementedError

    def single_forward(self, layers, blocks, h):
        for layer, block in zip(layers, blocks):
            h = layer(block, h)
        return h

    def forward(self, blocks, input_embeddings):
        '''
        Args:
            input_embeddings: {etype: {node_name: ...}}
        Returns:
            {etype: {node_name: ...}}
        '''
        different_embeddings = True if isinstance(
            list(input_embeddings.values())[0], dict) else False
        # dgl.add_self_loop(g) # TODO
        output_embeddings = {}
        for etype in self.primary_etypes:
            if different_embeddings:
                h = input_embeddings[etype]
            else:
                h = {
                    node_name: input_embeddings[node_name]
                    for node_name in [etype[0], etype[2]]
                }
            assert len(h) == 2
            etype_layers = self.layer_dict[str(etype)]
            # TODO
            if blocks[0].is_block:
                etype_blocks = blocks
            else:
                etype_blocks = [
                    dgl.edge_type_subgraph(
                        block, [etype, (etype[2], f'{etype[1]}-by', etype[0])])
                    for block in blocks
                ]
                # TODO very dirty!
                for x in etype_blocks:
                    x.is_block = True
            h = self.single_forward(etype_layers, etype_blocks, h)
            assert len(h) == 2, 'Unknown error'
            output_embeddings[etype] = h
        return output_embeddings


# if __name__ == '__main__':
#     import torch
#     from ....utils import add_reverse
#     graph = dgl.heterograph(
#         add_reverse({
#             ('user', 'follow', 'category'):
#             (torch.randint(10, (100, )), torch.randint(10, (100, ))),
#             ('user', 'buy', 'game'):
#             (torch.randint(10, (100, )), torch.randint(10, (100, ))),
#             ('user', 'play', 'game'): (torch.randint(10, (100, )),
#                                        torch.randint(10, (100, ))),
#             ('game', 'belong', 'category'): (torch.randint(10, (100, )),
#                                              torch.randint(10, (100, ))),
#         }), {
#             'user': 10,
#             'category': 10,
#             'game': 10
#         })
#     model = GCN([16, 12, 8, 6], graph.canonical_etypes)
#     print(model)

#     # Test full graph
#     input_embeddings1 = {
#         node_name: torch.rand(10, 16)
#         for node_name in ['user', 'category', 'game']
#     }
#     input_embeddings2 = {
#         etype:
#         {node_name: torch.rand(10, 16)
#          for node_name in [etype[0], etype[2]]}
#         for etype in graph.canonical_etypes if not etype[1].endswith('-by')
#     }
#     for x in [input_embeddings1, input_embeddings2]:
#         outputs = model([graph] * 3, x)
#         for k, v in outputs.items():
#             for k2, v2 in v.items():
#                 print(k, k2, v2.shape)

#     # Test mini-batch
#     eid_dict = {
#         etype: graph.edges('eid', etype=etype)
#         for etype in graph.canonical_etypes
#     }
#     dataloader = dgl.dataloading.EdgeDataLoader(
#         graph,
#         eid_dict,
#         dgl.dataloading.MultiLayerFullNeighborSampler(3),
#         batch_size=4,
#         shuffle=True,
#     )
#     for input_nodes, _, blocks in dataloader:
#         input_embeddings1 = {
#             k: torch.rand(v.size(0), 16)
#             for k, v in input_nodes.items()
#         }
#         input_embeddings2 = {
#             etype: {
#                 k: torch.rand(v.size(0), 16)
#                 for k, v in input_nodes.items() if k in [etype[0], etype[2]]
#             }
#             for etype in graph.canonical_etypes if not etype[1].endswith('-by')
#         }

#         for x in [input_embeddings1, input_embeddings2]:
#             outputs = model(blocks, x)
#             for k, v in outputs.items():
#                 for k2, v2 in v.items():
#                     print(k, k2, v2.shape)

#         break



import torch
import torch.nn as nn
import dgl

from .aggregator import GCN, LightGCN, GAT, NGCF
from ..general.attention import AdditiveAttention
from ..general.predictor import DNNPredictor, DotPredictor

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


class HeterogeneousNetwork(nn.Module):
    '''
    A general module for all graph-based models.
    '''
    def __init__(self, args, graph, tasks):
        super().__init__()
        self.args = args
        self.graph = graph
        self.primary_etypes = [
            x for x in graph.canonical_etypes if not x[1].endswith('-by')
        ]
        if args.different_embeddings:
            self.embedding = nn.ModuleDict({
                str(etype): nn.ModuleDict({
                    node_name: nn.Embedding(graph.num_nodes(node_name),
                                            args.graph_embedding_dims[0])
                    for node_name in [etype[0], etype[2]]
                })
                for etype in self.primary_etypes
            })
        else:
            self.embedding = nn.ModuleDict({
                node_name: nn.Embedding(graph.num_nodes(node_name),
                                        args.graph_embedding_dims[0])
                for node_name in graph.ntypes
            })

        if args.model_name in ['GCN', 'HET-GCN']:
            self.aggregator = GCN(args.graph_embedding_dims,
                                  graph.canonical_etypes)
        elif args.model_name in ['LightGCN', 'HET-LightGCN']:
            self.aggregator = LightGCN(args.graph_embedding_dims,
                                       graph.canonical_etypes)
        elif args.model_name in ['GAT', 'HET-GAT']:
            self.aggregator = GAT(args.graph_embedding_dims,
                                  graph.canonical_etypes,
                                  args.num_attention_heads)
        elif args.model_name in ['NGCF', 'HET-NGCF']:
            self.aggregator = NGCF(args.graph_embedding_dims,
                                   graph.canonical_etypes)
        else:
            raise NotImplementedError

        if args.model_name in ['GAT', 'HET-GAT']:
            final_single_embedding_dim = self.args.graph_embedding_dims[
                -1] * self.args.num_attention_heads
        elif args.model_name in ['LightGCN', 'HET-LightGCN']:
            final_single_embedding_dim = self.args.graph_embedding_dims[0]
        else:
            final_single_embedding_dim = self.args.graph_embedding_dims[-1]

        if args.embedding_aggregator == 'concat':
            embedding_num_dict = {
                node_name: sum([
                    node_name in [etype[0], etype[2]]
                    for etype in self.primary_etypes
                ])
                for node_name in graph.ntypes
            }
            final_embedding_dim_dict = {
                task['name']: (final_single_embedding_dim *
                               embedding_num_dict[task['scheme'][0]],
                               final_single_embedding_dim *
                               embedding_num_dict[task['scheme'][2]])
                for task in tasks
            }
        elif args.embedding_aggregator == 'attn':
            final_embedding_dim_dict = {
                task['name']:
                (final_single_embedding_dim, final_single_embedding_dim)
                for task in tasks
            }
            self.additive_attention = AdditiveAttention(
                args.attention_query_vector_dim, final_single_embedding_dim)
        else:
            raise NotImplementedError

        if args.predictor == 'dot':
            self.predictor = DotPredictor()
        elif args.predictor == 'dnn':
            self.predictor = nn.ModuleDict({
                task['name']:
                DNNPredictor(args.dnn_predictor_dims
                             if args.dnn_predictor_dims[0] != -1 else [
                                 sum(final_embedding_dim_dict[task['name']]
                                     ), *args.dnn_predictor_dims[1:]
                             ])
                for task in tasks
            })
        else:
            raise NotImplementedError

    def aggregate_embeddings(self, input_nodes, blocks):
        if self.args.different_embeddings:
            input_embeddings = {
                etype: {
                    node_name: self.embedding[str(etype)][node_name](
                        input_nodes[node_name])
                    for node_name in [etype[0], etype[2]]
                }
                for etype in self.primary_etypes
            }
        else:
            input_embeddings = {
                node_name: self.embedding[node_name](input_nodes[node_name])
                for node_name in self.graph.ntypes
            }

        output_embeddings = self.aggregator(blocks, input_embeddings)
        # transpose the nested dict
        output_embeddings = {
            node_name: [
                output_embeddings[etype][node_name]
                for etype in output_embeddings.keys()
                if node_name in output_embeddings[etype]
            ]
            for node_name in self.graph.ntypes
        }

        if self.args.embedding_aggregator == 'concat':

            def embedding_aggregator(x):
                return torch.cat(x, dim=-1)
        elif self.args.embedding_aggregator == 'attn':

            def embedding_aggregator(x):
                return self.additive_attention(torch.stack(x, dim=1))

        output_embeddings = {
            k: embedding_aggregator(v)
            for k, v in output_embeddings.items()
        }
        return output_embeddings

    def forward(self, first, second, task_name, provided_embeddings):
        '''
        Args:
            first: {
                'name': str,
                'index': (shape) batch_size
            },
            second: {
                'name': str,
                'index': (shape) batch_size
            }
        '''
        assert provided_embeddings is not None

        if isinstance(self.predictor, nn.ModuleDict):
            predictor = self.predictor[task_name]
        else:
            predictor = self.predictor

        return predictor(provided_embeddings[first['name']][first['index']],
                         provided_embeddings[second['name']][second['index']])


if __name__ == '__main__':
    graph = dgl.heterograph(
        {
            ('user', 'follows', 'user'):
            (torch.randint(10, (20, )), torch.randint(10, (20, ))),
            ('user', 'follows', 'topic'):
            (torch.randint(10, (20, )), torch.randint(10, (20, ))),
            ('user', 'plays', 'game'):
            (torch.randint(10, (20, )), torch.randint(10, (20, )))
        }, {
            'user': 10,
            'topic': 10,
            'game': 10
        })
    graph = dgl.to_simple(graph)
    graph = graph.to(device)
    from parameters import parse_args
    args = parse_args()
    for x in ['HET-GCN', 'HET-GAT']:
        args.model_name = x
        model = HeterogeneousNetwork(args, graph).to(device)
        first_index = torch.randint(10, (64, )).to(device)
        second_index = torch.randint(10, (64, )).to(device)
        first = {'name': 'user', 'index': first_index}
        second = {'name': 'topic', 'index': second_index}
        y_pred = model(first, second)
        print(y_pred)



import torch
import dgl

from .lightgcn_conv import LightGCNConv
from ..base import HeterogeneousAggregator


def get_index(a, b):
    '''
    For each value in b, get its index in a and return the indexs
    Args:
        a: 1D tensor
        b: 1D tensor
    Returns:
        indexs (the same shape as b)
    '''
    assert len(a.size()) == 1
    assert len(b.size()) == 1
    index = a.argsort()
    return index[torch.bucketize(b, a[index])]


class LightGCN(HeterogeneousAggregator):
    def __init__(self, graph_embedding_dims, etypes):
        super().__init__(graph_embedding_dims, etypes)

    def get_layer(self, input_dim, output_dim, current_layer, total_layer):
        assert 0 <= current_layer <= total_layer - 1

        return LightGCNConv()

    # TODO
    # I found that adding this (i.e., mean over all layer outputs) contributes to a lower score.
    # Maybe there is something wrong with my implementations.
    # Anyway, I commented out this to make its default behavior take effect (i.e. use only the last layer output)
    # for a better performance

    # def single_forward(self, layers, blocks, h):
    #     outputs = []
    #     for layer, block in zip(layers, blocks):
    #         if len(block.srcdata[dgl.NID]) == 0:
    #             # If is not a block.
    #             outputs.append(h)
    #         else:
    #             # Since blocks have different size of input and output nodes,
    #             # We should only save the nodes available in the last block output
    #             outputs.append({
    #                 k: v[get_index(block.srcdata[dgl.NID][k],
    #                                blocks[-1].dstdata[dgl.NID][k])]
    #                 for k, v in h.items()
    #             })
    #         h = layer(block, h)
    #     outputs.append(h)
    #     outputs = {
    #         k: torch.mean(torch.stack([x[k] for x in outputs], dim=-1), dim=-1)
    #         for k in outputs[0].keys()
    #     }
    #     return outputs



import torch.nn as nn
import dgl
from dgl.nn.pytorch import GATConv

from .base import HeterogeneousAggregator


class GAT(HeterogeneousAggregator):
    def __init__(self, graph_embedding_dims, etypes, num_attention_heads):
        self.num_attention_heads = num_attention_heads  # Must put this before super().__init__()
        super().__init__(graph_embedding_dims, etypes)

    def get_layer(self, input_dim, output_dim, current_layer, total_layer):
        assert 0 <= current_layer <= total_layer - 1
        if current_layer < total_layer - 1:
            return GATConv(
                input_dim *
                (self.num_attention_heads if current_layer >= 1 else 1),
                output_dim,
                self.num_attention_heads,
                activation=nn.ELU())

        return GATConv(
            input_dim * (self.num_attention_heads if total_layer >= 2 else 1),
            output_dim, self.num_attention_heads)

    def single_forward(self, layers, blocks, h):
        for layer, block in zip(layers, blocks):
            h = layer(block, h)
            h = {k: v.view(v.size(0), -1) for k, v in h.items()}

        return h


if __name__ == '__main__':
    import torch
    graph = dgl.graph((torch.randint(10, (20, )), torch.randint(10, (20, ))))
    graph = dgl.to_simple(graph)
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    graph = graph.to(device)
    model = GAT([8, 6], 8).to(device)
    inputs = torch.rand(10, 8).to(device)
    print(model(graph, inputs))
    model = GAT([16, 12, 8, 6], 8).to(device)
    inputs = torch.rand(10, 16).to(device)
    print(model(graph, inputs))



import argparse
from distutils.util import strtobool


def str2bool(x):
    return bool(strtobool(x))


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--num_epochs', type=int, default=1000)
    parser.add_argument('--learning_rate', type=float, default=0.0005)
    parser.add_argument('--batch_size', type=int, default=4096)
    parser.add_argument('--num_workers', type=int, default=16)
    parser.add_argument('--non_graph_embedding_dim', type=int, default=200)
    parser.add_argument('--graph_embedding_dims',
                        type=int,
                        nargs='+',
                        default=[200, 128, 64])
    parser.add_argument(
        '--neighbors_sampling_quantile',
        type=float,
        default=0.9,
        help=
        'Set the number of sampled neighbors to the quantile of the numbers of neighbors'
    )
    parser.add_argument('--min_neighbors_sampled', type=int, default=4)
    parser.add_argument('--max_neighbors_sampled', type=int, default=512)
    parser.add_argument('--single_attribute_dim', type=int,
                        default=40)  # TODO: support attributes
    parser.add_argument('--attention_query_vector_dim', type=int, default=200)
    parser.add_argument(
        '--dnn_predictor_dims',
        type=int,
        nargs='+',
        default=[-1, 128, 1],
        help=
        'You can set first dim as -1 to make it automatically fit the input vector'
    )
    parser.add_argument('--num_batches_show_loss', type=int, default=50)
    parser.add_argument('--num_epochs_validate', type=int, default=5)
    parser.add_argument('--early_stop_patience', type=int, default=20)
    parser.add_argument('--num_attention_heads', type=int, default=8)
    parser.add_argument('--save_checkpoint', type=str2bool, default=False)
    parser.add_argument('--different_embeddings', type=str2bool, default=False)
    parser.add_argument('--negative_sampling_ratio', type=int, default=4)

    parser.add_argument(
        '--model_name',
        type=str,
        default='GCN',
        choices=[
            # Non-graph
            'NCF',

            # Graph with single type of edge (we think it as homogeneous graph)
            'GCN',
            'GAT',
            'LightGCN',
            'NGCF',

            # Graph with multiple types of edge (we think it as heterogeneous graph)
            'HET-GCN',
            'HET-GAT',
            'HET-NGCF',
            'HET-LightGCN',

            # To be categorized
            'GraphRec',
            'DeepFM',
            'DSSM',
            'DiffNet',
            'DiffNet++',
            'DANSER'
        ])
    parser.add_argument('--dataset', type=str, default='jd')
    parser.add_argument('--embedding_aggregator',
                        type=str,
                        default='concat',
                        choices=['concat', 'attn'])
    parser.add_argument('--predictor',
                        type=str,
                        default='dnn',
                        choices=['dot', 'dnn'])
    parser.add_argument('--metadata_path', type=str)
    parser.add_argument('--edge_choice',
                        type=int,
                        nargs='+',
                        default=[],
                        help='Left empty to use all in metadata file')
    parser.add_argument('--training_task_choice',
                        type=int,
                        nargs='+',
                        default=[],
                        help='Left empty to use all in metadata file')
    parser.add_argument('--evaluation_task_choice',
                        type=int,
                        nargs='+',
                        default=[],
                        help='Left empty to use all in `training_task_choice`')
    parser.add_argument('--task_loss_overwrite', type=str, nargs='+')
    parser.add_argument('--task_weight_overwrite', type=float, nargs='+')
    args, unknown = parser.parse_known_args()
    if len(unknown) > 0:
        print(
            'Warning: if you are not in testing mode, you may have got some parameters wrong input'
        )
    if args.metadata_path is None:
        args.metadata_path = f'metadata/{args.dataset}.json'
    return args



import numpy as np


class EarlyStopping:
    def __init__(self, patience=5):
        self.patience = patience
        self.counter = 0
        self.best_loss = np.Inf

    def __call__(self, val_loss):
        """
        if you use other metrics where a higher value is better, e.g. accuracy,
        call this with its corresponding negative value
        """
        if val_loss < self.best_loss:
            early_stop = False
            get_better = True
            self.counter = 0
            self.best_loss = val_loss
        else:
            get_better = False
            self.counter += 1
            if self.counter >= self.patience:
                early_stop = True
            else:
                early_stop = False

        return early_stop, get_better



import torch
import json

from .utils import evaluate, latest_checkpoint, create_model, create_logger, process_metadata, dict2table
from .parameters import parse_args

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
args = parse_args()


def test():
    with open(args.metadata_path) as f:
        metadata = json.load(f)
        metadata = process_metadata(metadata)
        logger.info(metadata)

    model = create_model(metadata, logger).to(device)
    test_metrics_dict = {}
    for task in metadata['task']:
        if not task['evaluation']:
            continue

        checkpoint_path = latest_checkpoint(
            f"./checkpoint/{args.model_name}-{args.dataset}/{task['name']}")
        if checkpoint_path is None:
            logger.error(f"No checkpoint file for task {task['name']} found!")
            exit()
        logger.info(f'Load saved parameters in {checkpoint_path}')
        checkpoint = torch.load(checkpoint_path)
        model.load_state_dict(checkpoint['model_state_dict'])
        model.eval()
        metrics, _ = evaluate(model, [task], 'test')
        test_metrics_dict[task['name']] = metrics[task['name']]

    logger.info(f'Metrics on test set\n{dict2table(test_metrics_dict)}')


if __name__ == '__main__':
    logger = create_logger()
    logger.info(args)
    logger.info(f'Using device: {device}')
    logger.info(f'Testing model {args.model_name} with dataset {args.dataset}')
    test()



import torch
import torch.nn as nn
import dgl.function as fn
from dgl.utils import expand_as_pair


class LightGCNConv(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, graph, feature):
        with graph.local_scope():
            feat_src, feat_dst = expand_as_pair(feature, graph)

            degs = graph.out_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1, ) * (feat_src.dim() - 1)
            norm = torch.reshape(norm, shp)
            feat_src = feat_src * norm

            graph.srcdata['h'] = feat_src

            graph.update_all(fn.copy_u('h', 'm'), fn.sum('m', 'h'))
            rst = graph.dstdata['h']

            degs = graph.in_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1, ) * (feat_dst.dim() - 1)
            norm = torch.reshape(norm, shp)
            rst = rst * norm

            return rst



import torch
import torch.nn as nn
import dgl
import numpy as np
import pandas as pd
import json
import os
import time
import datetime
import enlighten
import copy
from torch.utils.tensorboard import SummaryWriter

from .parameters import parse_args
from .utils import EarlyStopping, evaluate, time_since, create_model, create_logger, is_graph_model, process_metadata, dict2table, deep_apply, is_single_relation_model
from .loss import BPRLoss, MarginLoss

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
args = parse_args()


def train():
    with open(args.metadata_path) as f:
        metadata = json.load(f)
        metadata = process_metadata(metadata)
        logger.info(metadata)

    model = create_model(metadata, logger).to(device)
    logger.info(model)

    task_to_evaluate = [x['name'] for x in metadata['task'] if x['evaluation']]

    model.eval()
    metrics, _ = evaluate(
        model, [x for x in metadata['task'] if x['name'] in task_to_evaluate],
        'val')
    model.train()
    logger.info(f'Initial metrics on validation set {deep_apply(metrics)}')
    best_checkpoint_dict = {
        task: copy.deepcopy(model.state_dict())
        for task in task_to_evaluate
    }
    best_val_metrics_dict = copy.deepcopy(metrics)

    criterions = {}

    if is_single_relation_model():
        assert len(metadata['task']) == 1

    for task in metadata['task']:
        # criterions
        if task['type'] == 'top-k-recommendation':
            original_loss_map = {
                'NGCF': 'bpr',
                'HET-NGCF': 'bpr',
                # TODO
            }
            if args.model_name in original_loss_map and task[
                    'loss'] != original_loss_map[args.model_name]:
                logger.warning(
                    'You are using a different type of loss with the type in the paper'
                )
            if task['loss'] == 'binary-cross-entropy':
                criterions[task['name']] = nn.BCEWithLogitsLoss()
            elif task['loss'] == 'cross-entropy':
                criterions[task['name']] = nn.CrossEntropyLoss()
            elif task['loss'] == 'bpr':
                criterions[task['name']] = BPRLoss()
            elif task['loss'] == 'margin':
                criterions[task['name']] = MarginLoss()
            else:
                raise NotImplementedError

        elif task['type'] == 'interaction-attribute-regression':
            raise NotImplementedError
        else:
            raise NotImplementedError

    optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)
    loss_full = []
    early_stopping_dict = {
        task['name']: EarlyStopping(args.early_stop_patience)
        for task in metadata['task']
    }
    start_time = time.time()
    writer = SummaryWriter(
        log_dir=
        f"./runs/{args.model_name}-{args.dataset}/{str(datetime.datetime.now().replace(microsecond=0)).replace(' ', '_').replace(':', '-')}{'-remark-' + os.environ['REMARK'] if 'REMARK' in os.environ else ''}"
    )

    if args.save_checkpoint:
        for task in metadata['task']:
            os.makedirs(
                f"./checkpoint/{args.model_name}-{args.dataset}/{task['name']}",
                exist_ok=True)

    enlighten_manager = enlighten.get_manager()

    batch = 0
    if is_graph_model():
        etype2num_neighbors = {
            etype: np.clip(
                np.quantile(model.graph.in_degrees(etype=etype),
                            args.neighbors_sampling_quantile,
                            interpolation='nearest'),
                args.min_neighbors_sampled, args.max_neighbors_sampled)
            for etype in model.graph.canonical_etypes
        }
        logger.debug(f'Neighbors sampled {etype2num_neighbors}')

    try:
        with enlighten_manager.counter(total=args.num_epochs,
                                       desc='Training epochs',
                                       unit='epochs') as epoch_pbar:
            for epoch in epoch_pbar(range(1, args.num_epochs + 1)):
                if is_graph_model():
                    neighbor_sampler = dgl.dataloading.MultiLayerNeighborSampler(
                        [etype2num_neighbors] *
                        (len(args.graph_embedding_dims) - 1))
                else:
                    neighbor_sampler = dgl.dataloading.MultiLayerNeighborSampler(
                        [0])  # TODO

                # Two versions of edge_sampling, either is OK
                def edge_sampling(etype):
                    df = pd.DataFrame(
                        torch.stack(model.graph.edges(etype=etype),
                                    dim=1).numpy())
                    return df.sample(frac=1).drop_duplicates(0).index.values

                # def edge_sampling(etype):
                #     subgraph = dgl.edge_type_subgraph(model.graph, [etype])
                #     return dgl.sampling.sample_neighbors(
                #         subgraph, {
                #             etype[0]: subgraph.nodes(etype[0])
                #         },
                #         1,
                #         edge_dir='out').edata[dgl.EID]

                eid_dict = {
                    etype: edge_sampling(etype)
                    for etype in
                    model.primary_etypes  # TODO model.graph.canonical_etypes ?
                }

                # parse reverse_etypes
                etypes = copy.deepcopy(model.graph.canonical_etypes)
                reverse_etypes = {}
                for etype in model.graph.canonical_etypes:
                    if etype[1].endswith('-by'):
                        reverse_etype = (etype[2], etype[1][:-len('-by')],
                                         etype[0])
                        reverse_etypes[etype] = reverse_etype
                        reverse_etypes[reverse_etype] = etype
                        etypes.remove(etype)
                        etypes.remove(reverse_etype)
                assert len(etypes) == 0

                dataloader = dgl.dataloading.EdgeDataLoader(
                    model.graph,
                    eid_dict,
                    neighbor_sampler,
                    exclude='reverse_types',
                    reverse_etypes=reverse_etypes,
                    negative_sampler=dgl.dataloading.negative_sampler.Uniform(
                        args.negative_sampling_ratio),
                    batch_size=args.batch_size,
                    shuffle=True,
                    drop_last=False,
                    num_workers=args.num_workers,
                    pin_memory=True)

                with enlighten_manager.counter(total=len(dataloader),
                                               desc='Training batches',
                                               unit='batches',
                                               leave=False) as batch_pbar:
                    for input_nodes, positive_graph, negative_graph, blocks in batch_pbar(
                            dataloader):
                        batch += 1
                        if is_graph_model():
                            if batch == 1:
                                node_coverage = {
                                    k: len(v) / model.graph.num_nodes(k)
                                    for k, v in input_nodes.items()
                                }
                                logger.debug(
                                    f'Node coverage {deep_apply(node_coverage)}'
                                )
                            input_nodes = {
                                k: v.to(device)
                                for k, v in input_nodes.items()
                            }
                        positive_graph = positive_graph.to(device)
                        negative_graph = negative_graph.to(device)
                        if is_graph_model():
                            blocks = [block.to(device) for block in blocks]
                            output_embeddings = model.aggregate_embeddings(
                                input_nodes, blocks)
                        else:
                            output_embeddings = None
                        loss = 0
                        for task in metadata['task']:
                            positive_index = torch.stack(
                                positive_graph.edges(etype=task['scheme']))
                            negative_index = torch.stack(
                                negative_graph.edges(etype=task['scheme']))
                            index = torch.cat((positive_index, negative_index),
                                              dim=1)
                            if not is_graph_model():
                                # map indexs
                                index[0] = positive_graph.ndata[dgl.NID][
                                    task['scheme'][0]][index[0]]
                                index[1] = positive_graph.ndata[dgl.NID][
                                    task['scheme'][2]][index[1]]
                            first = {
                                'name': task['scheme'][0],
                                'index': index[0]
                            }
                            second = {
                                'name': task['scheme'][2],
                                'index': index[1]
                            }
                            y_pred = model(first, second, task['name'],
                                           output_embeddings)
                            if task['loss'] == 'binary-cross-entropy':
                                y_true = torch.cat(
                                    (torch.ones(positive_index.size(1)),
                                     torch.zeros(
                                         negative_index.size(1)))).to(device)
                                task_loss = criterions[task['name']](y_pred,
                                                                     y_true)
                            elif task['loss'] == 'cross-entropy':
                                assert torch.equal(
                                    positive_index[0].expand(
                                        args.negative_sampling_ratio,
                                        -1).transpose(0, 1),
                                    negative_index[0].view(
                                        -1, args.negative_sampling_ratio))
                                sample_length = positive_index.size(1)
                                positive_pred = y_pred[:
                                                       sample_length].unsqueeze(
                                                           dim=-1)
                                negative_pred = y_pred[sample_length:].view(
                                    sample_length,
                                    args.negative_sampling_ratio)
                                y_pred = torch.cat(
                                    (positive_pred, negative_pred), dim=1)
                                y_true = torch.zeros(sample_length).long().to(
                                    device)
                                task_loss = criterions[task['name']](y_pred,
                                                                     y_true)
                            elif task['loss'] == 'margin':
                                assert torch.equal(
                                    positive_index[0].expand(
                                        args.negative_sampling_ratio,
                                        -1).transpose(0, 1),
                                    negative_index[0].view(
                                        -1, args.negative_sampling_ratio))
                                sample_length = positive_index.size(1)
                                positive_pred = y_pred[:sample_length]
                                negative_pred = y_pred[sample_length:]
                                task_loss = criterions[task['name']](
                                    positive_pred, negative_pred)
                            else:
                                raise NotImplementedError

                            loss += task_loss * task['weight']

                            if len(metadata['task']) > 1:
                                writer.add_scalar(f"Train/Loss/{task['name']}",
                                                  task_loss.item(), batch)

                        optimizer.zero_grad()
                        loss.backward()
                        optimizer.step()
                        loss_full.append(loss.item())
                        writer.add_scalar('Train/Loss', loss.item(), batch)
                        if batch % args.num_batches_show_loss == 0:
                            logger.info(
                                f"Time {time_since(start_time)}, epoch {epoch}, batch {batch}, current loss {loss.item():.4f}, average loss {np.mean(loss_full):.4f}, latest average loss {np.mean(loss_full[-10:]):.4f}"
                            )

                if epoch % args.num_epochs_validate == 0:
                    model.eval()
                    metrics, overall = evaluate(model, [
                        x for x in metadata['task']
                        if x['name'] in task_to_evaluate
                    ], 'val')
                    model.train()

                    for task_name, values in metrics.items():
                        for metric, value in values.items():
                            writer.add_scalar(
                                f'Validation/{task_name}/{metric}', value,
                                epoch)
                    logger.info(
                        f"Time {time_since(start_time)}, epoch {epoch}, metrics {deep_apply(metrics)}"
                    )
                    for task_name in copy.deepcopy(task_to_evaluate):
                        early_stop, get_better = early_stopping_dict[
                            task_name](-overall[task_name])
                        if early_stop:
                            task_to_evaluate.remove(task_name)
                            logger.info(f'Task {task_name} early stopped')
                        elif get_better:
                            best_checkpoint_dict[task_name] = copy.deepcopy(
                                model.state_dict())
                            best_val_metrics_dict[task_name] = copy.deepcopy(
                                metrics[task_name])
                            if args.save_checkpoint:
                                torch.save(
                                    {'model_state_dict': model.state_dict()},
                                    f"./checkpoint/{args.model_name}-{args.dataset}/{task_name}/ckpt-{epoch}.pt"
                                )

                    if not task_to_evaluate:
                        logger.info('All tasks early stopped')
                        break

    except KeyboardInterrupt:
        logger.info('Stop in advance')

    logger.info(
        f'Best metrics on validation set\n{dict2table(best_val_metrics_dict)}')
    test_metrics_dict = {}
    for task_name, checkpoint in best_checkpoint_dict.items():
        model.load_state_dict(checkpoint)
        model.eval()
        metrics, _ = evaluate(
            model, [x for x in metadata['task'] if x['name'] == task_name],
            'test')
        test_metrics_dict[task_name] = metrics[task_name]
    logger.info(f'Metrics on test set\n{dict2table(test_metrics_dict)}')


if __name__ == '__main__':
    logger = create_logger()
    logger.info(args)
    logger.info(f'Using device: {device}')
    logger.info(
        f'Training model {args.model_name} with dataset {args.dataset}')
    train()



import numpy as np
import pandas as pd
import time
import dgl
import torch
import os
import logging
import coloredlogs
import math
import datetime
import copy
from itertools import chain

from .metrics import *
from ..model import *
from ..parameters import parse_args

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
args = parse_args()

# A simple cache mechanism for df reading and sorting, since it will be run for many times
_df_cache_for_validation = {}


@torch.no_grad()
def evaluate(model, tasks, mode):
    metrics = {}

    if is_graph_model():
        input_nodes = {
            node_name: model.graph.nodes(ntype=node_name).to(device)
            for node_name in model.graph.ntypes
        }
        provided_embeddings = model.aggregate_embeddings(
            input_nodes,
            [model.graph.to(device)] * (len(args.graph_embedding_dims) - 1))
    else:
        provided_embeddings = None

    for task in tasks:
        file_path = f"./data/{args.dataset}/{mode}/{task['filename']}"
        if mode == 'val' and file_path in _df_cache_for_validation:
            df = _df_cache_for_validation[file_path]
        else:
            df = pd.read_table(file_path)
            df.sort_values(df.columns[0], inplace=True)
            if mode == 'val':
                _df_cache_for_validation[file_path] = df

        columns = df.columns.tolist()
        test_data = np.transpose(df.values)
        test_data = torch.from_numpy(test_data).to(device)
        first_indexs, second_indexs, y_trues = test_data
        y_preds = []
        y_trues = y_trues.cpu().numpy()
        for i in range(math.ceil(len(df) / (8 * args.batch_size))):
            first_index = first_indexs[i * (8 * args.batch_size):(i + 1) *
                                       (8 * args.batch_size)]
            second_index = second_indexs[i * (8 * args.batch_size):(i + 1) *
                                         (8 * args.batch_size)]
            first = {'name': columns[0], 'index': first_index}
            second = {'name': columns[1], 'index': second_index}
            y_pred = model(first, second, task['name'], provided_embeddings)
            y_pred = y_pred.cpu().numpy()
            y_preds.append(y_pred)

        y_preds = np.concatenate(y_preds, axis=0)

        if task['type'] == 'top-k-recommendation':
            single_sample_length = df.groupby(columns[0]).size().values
            assert len(
                set(single_sample_length)
            ) == 1, f'The number of {columns[1]}s for different {columns[0]}s should be equal'
            y_trues = y_trues.reshape(-1, single_sample_length[0])
            y_preds = y_preds.reshape(-1, single_sample_length[0])
            metrics[task['name']] = {
                'AUC':
                fast_roc_auc_score(y_trues,
                                   y_preds,
                                   num_processes=args.num_workers),
                'MRR':
                mrr(y_trues, y_preds),
                'NDCG@10':
                ndcg_score(y_trues, y_preds, k=10, ignore_ties=True),
                'NDCG@50':
                ndcg_score(y_trues, y_preds, k=50, ignore_ties=True),
                'Recall@10':
                recall(y_trues, y_preds, k=10),
                'Recall@50':
                recall(y_trues, y_preds, k=50),
            }
        elif task['type'] == 'interaction-attribute-regression':
            raise NotImplementedError
        else:
            raise NotImplementedError

    overall = {k: np.mean(list(v.values())) for k, v in metrics.items()}
    return metrics, overall


def time_since(since):
    """
    Format elapsed time string.
    """
    now = time.time()
    elapsed_time = now - since
    return time.strftime("%H:%M:%S", time.gmtime(elapsed_time))


def latest_checkpoint(directory):
    if not os.path.exists(directory):
        return None
    all_checkpoints = {
        int(x.split('.')[-2].split('-')[-1]): x
        for x in os.listdir(directory) if 'keep' not in x
    }
    if not all_checkpoints:
        return None
    return os.path.join(directory,
                        all_checkpoints[max(all_checkpoints.keys())])


def create_model(metadata, logger):
    num_nodes_dict = {
        node['name']:
        len(pd.read_table(f"./data/{args.dataset}/train/{node['name']}.tsv"))
        for node in metadata['graph']['node']
    }
    for node in metadata['graph']['node']:
        if len(node['attribute']) != 0:
            logger.warning(
                f"The attributes of node {node['name']} are ignored")

    if is_single_relation_model():
        assert len(metadata['graph']['node']) == 2
        assert len(metadata['graph']['edge']) == 1
        first_node_name = metadata['graph']['edge'][0]['scheme'][0]
        second_node_name = metadata['graph']['edge'][0]['scheme'][2]
    else:
        assert len(metadata['graph']['edge']) > 1

    # TODO
    # if args.model_name == 'HET-GraphRec':
    #     assert len(metadata['graph']['edge']) == 2
    #     node_names = [
    #         metadata['graph']['edge'][0]['scheme'][0],
    #         metadata['graph']['edge'][0]['scheme'][2],
    #         metadata['graph']['edge'][1]['scheme'][0],
    #         metadata['graph']['edge'][1]['scheme'][2],
    #     ]
    #     assert node_names.count('user') == 3 and node_names.count(
    #         'item') == 3

    graph_data = {}

    for edge in metadata['graph']['edge']:
        if edge['scheme'][0] == edge['scheme'][2]:
            raise NotImplementedError

        df = pd.read_table(f"./data/{args.dataset}/train/{edge['filename']}")
        graph_data[edge['scheme']] = (torch.as_tensor(df.iloc[:, 0].values),
                                      torch.as_tensor(df.iloc[:, 1].values))

    graph = dgl.heterograph(add_reverse(graph_data), num_nodes_dict)
    if is_graph_model():
        for edge in metadata['graph']['edge']:
            if edge['weighted']:
                raise NotImplementedError
        model = HeterogeneousNetwork(args, graph, metadata['task'])
        return model

    if args.model_name == 'NCF':
        model = NCF(args, graph, num_nodes_dict[first_node_name],
                    num_nodes_dict[second_node_name])
        return model

    raise NotImplementedError(
        f'This model {args.model_name} is under development')


def create_logger():
    logger = logging.getLogger(__name__)
    coloredlogs.install(level='DEBUG',
                        logger=logger,
                        fmt='%(asctime)s %(levelname)s %(message)s')
    log_dir = f'./log/{args.model_name}-{args.dataset}'
    os.makedirs(log_dir, exist_ok=True)
    log_file_path = os.path.join(
        log_dir,
        f"{str(datetime.datetime.now()).replace(' ', '_').replace(':', '-')}{'-remark-' + os.environ['REMARK'] if 'REMARK' in os.environ else ''}.txt"
    )
    logger.info(f'Check {log_file_path} for the log of this run')
    file_hander = logging.FileHandler(log_file_path)
    logger.addHandler(file_hander)
    return logger


def is_graph_model():
    '''
    Whether need the message passing on the graph
    '''
    if args.model_name in ['NCF']:
        return False
    if args.model_name in [
            'GCN', 'LightGCN', 'GAT', 'NGCF', 'HET-GCN', 'HET-LightGCN',
            'HET-GAT', 'HET-NGCF', 'GraphRec'
    ]:
        return True

    raise NotImplementedError


def is_single_relation_model():
    if args.model_name in ['NCF', 'GCN', 'LightGCN', 'GAT', 'NGCF']:
        return True
    if args.model_name in ['HET-GCN', 'HET-LightGCN', 'HET-GAT', 'HET-NGCF']:
        return False

    raise NotImplementedError


def process_metadata(metadata):
    def parse_scheme_from_filename(filename):
        filename = filename.split('.')[0].split('-')
        assert len(filename) == 3
        return tuple(filename[x] for x in [0, 2, 1])

    for edge in metadata['graph']['edge']:
        edge['scheme'] = parse_scheme_from_filename(edge['filename'])
    for task in metadata['task']:
        task['scheme'] = parse_scheme_from_filename(task['filename'])
        task['name'] = os.path.splitext(task['filename'])[0]

    if args.edge_choice:
        metadata['graph']['edge'] = [
            metadata['graph']['edge'][x] for x in args.edge_choice
        ]
    node_from_edge = set(
        chain.from_iterable([[edge['scheme'][0], edge['scheme'][2]]
                             for edge in metadata['graph']['edge']]))
    metadata['graph']['node'] = [
        x for x in metadata['graph']['node'] if x['name'] in node_from_edge
    ]

    training_task_choice = args.training_task_choice if args.training_task_choice else list(
        range(len(metadata['task'])))
    evaluation_task_choice = args.evaluation_task_choice if args.evaluation_task_choice else training_task_choice
    assert set(evaluation_task_choice) <= set(
        training_task_choice
    ), 'There are tasks in evaluation but not in training'
    metadata['task'] = [{
        **metadata['task'][x], 'evaluation': (x in evaluation_task_choice)
    } for x in training_task_choice]

    if args.task_loss_overwrite is not None:
        assert len(metadata['task']) == len(args.task_loss_overwrite)
        for task, loss in zip(metadata['task'], args.task_loss_overwrite):
            task['loss'] = loss

    if args.task_weight_overwrite is not None:
        assert len(metadata['task']) == len(args.task_weight_overwrite)
        for task, weight in zip(metadata['task'], args.task_weight_overwrite):
            task['weight'] = weight

    assert any([x['weight'] > 0 for x in metadata['task']
                ]), 'Make sure at least one task with positive weight'

    assert set([task['filename'] for task in metadata['task']]) <= set([
        edge['filename'] for edge in metadata['graph']['edge']
    ]), 'There are files in task metadata but not in graph edge metadata'

    return metadata


def add_reverse(graph_data):
    '''
    Add reverse edges for graph data before feed into `dgl.heterograph`
    '''
    for scheme in list(graph_data.keys()):
        if scheme[0] == scheme[2]:
            graph_data[scheme] = (
                torch.cat(graph_data[scheme]),
                torch.cat(graph_data[scheme][::-1]),
            )
        else:
            reversed_scheme = (scheme[2], f'{scheme[1]}-by', scheme[0])
            graph_data[reversed_scheme] = (graph_data[scheme][1],
                                           graph_data[scheme][0])
    return graph_data


def copy_arguments(f):
    def selectively_copy(x):
        if isinstance(x, list) or isinstance(x, dict):
            return copy.deepcopy(x)
        else:
            return x

    def wrapper(*args, **kwargs):
        args = tuple(selectively_copy(x) for x in args)
        kwargs = {k: selectively_copy(v) for k, v in kwargs.items()}
        return f(*args, **kwargs)

    return wrapper


@copy_arguments
def deep_apply(d, f=lambda x: f'{x:.4f}'):
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = deep_apply(v, f)
        else:
            d[k] = f(v)
    return d


def dict2table(d, k_fn=str, v_fn=lambda x: f'{x:.4f}'):
    '''
    Convert a nested dict to markdown table
    '''
    def parse_header(d, depth=0):
        if isinstance(list(d.values())[0], dict):
            header = parse_header(list(d.values())[0], depth=depth + 1)
            for v in d.values():
                assert header == parse_header(v, depth=depth + 1)
            return header
        else:
            return f"| {' | '.join([''] * depth + list(map(k_fn, d.keys())))} |"

    def parse_content(d, accumulated_keys=[]):
        if isinstance(list(d.values())[0], dict):
            contents = []
            for k, v in d.items():
                contents.extend(parse_content(v, accumulated_keys + [k_fn(k)]))
            return contents
        else:
            return [
                f"| {' | '.join(accumulated_keys + list(map(v_fn, d.values())))} |"
            ]

    lines = [parse_header(d), *parse_content(d)]
    return '\n'.join(lines)



import torch
import json

from .utils import latest_checkpoint, create_model, create_logger, process_metadata, is_graph_model
from .parameters import parse_args

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
args = parse_args()


@torch.no_grad()
def export():
    with open(args.metadata_path) as f:
        metadata = json.load(f)
        metadata = process_metadata(metadata)
        logger.info(metadata)

    model = create_model(metadata, logger).to(device)
    checkpoint_path = latest_checkpoint(
        f'./checkpoint/{args.model_name}-{args.dataset}')
    if checkpoint_path is None:
        logger.error('No checkpoint file found!')
        exit()
    logger.info(f'Load saved parameters in {checkpoint_path}')
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint['model_state_dict'])
    model.eval()
    if is_graph_model():
        input_nodes = {
            node_name: model.graph.nodes(ntype=node_name).to(device)
            for node_name in model.graph.ntypes
        }
        embeddings = model.aggregate_embeddings(
            input_nodes,
            [model.graph.to(device)] * (len(args.graph_embedding_dims) - 1))
        import ipdb
        ipdb.set_trace()
    else:
        raise NotImplementedError


if __name__ == '__main__':
    logger = create_logger()
    logger.info(args)
    logger.info(f'Using device: {device}')
    logger.info(
        f'Exporting embeddings for model {args.model_name} with dataset {args.dataset}'
    )
    export()



import torch
import torch.nn as nn

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


class DeepFM(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args

    def forward(self):
        pass



import torch.nn as nn
import dgl

from .ngcf_conv import NGCFConv
from ..base import HeterogeneousAggregator


class NGCF(HeterogeneousAggregator):
    def __init__(self, graph_embedding_dims, etypes):
        super().__init__(graph_embedding_dims, etypes)

    def get_layer(self, input_dim, output_dim, current_layer, total_layer):
        assert 0 <= current_layer <= total_layer - 1
        if current_layer < total_layer - 1:
            return NGCFConv(input_dim, output_dim, activation=nn.LeakyReLU())

        return NGCFConv(input_dim, output_dim)

    # TODO NGCF should concat the outputs of all layers
    # def single_forward(self, layers, blocks, h):
    #     outputs = []
    #     outputs.append(h)
    #     for layer, block in zip(layers, blocks):
    #         h = layer(block, h)
    #         outputs.append(h)
    #     return torch.cat(outputs, dim=-1)


if __name__ == '__main__':
    import torch
    graph = dgl.graph((torch.randint(10, (20, )), torch.randint(10, (20, ))))
    graph = dgl.to_simple(graph)
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    graph = graph.to(device)
    model = NGCF([16, 12, 8, 6]).to(device)
    print(model)
    inputs = torch.rand(10, 16).to(device)
    print(model(graph, inputs))
